'use strict';

/**
 * @fileOverview Snippets implementation.
 */

const singleCharacterEscapes = new Map([
  ['n', '\n'], ['r', '\r'], ['t', '\t'],
]);

const executableCode = new Map();

let snippetsLibrarySource = '';

/**
 * Parses a script and returns a list of all its commands and their arguments
 * @param {string} script
 * @return {Array.<string[]>}
 */
function parseScript(script) {
  const tree = [];

  let escape = false;
  let withinQuotes = false;

  let unicodeEscape = null;

  let call = [];
  let argument = '';

  for (const character of `${script.trim()};`) {
    if (unicodeEscape !== null) {
      unicodeEscape += character;

      if (unicodeEscape.length === 4) {
        const codePoint = parseInt(unicodeEscape, 16);
        if (!Number.isNaN(codePoint)) {
          argument += String.fromCodePoint(codePoint);
        }

        unicodeEscape = null;
      }
    } else if (escape) {
      escape = false;

      if (character === 'u') {
        unicodeEscape = '';
      } else {
        argument += singleCharacterEscapes.get(character) || character;
      }
    } else if (character === '\\') {
      escape = true;
    } else if (character === "'") {
      withinQuotes = !withinQuotes;
    } else if (withinQuotes || (character !== ';' && !/\s/.test(character))) {
      argument += character;
    } else {
      if (argument) {
        call.push(argument);
        argument = '';
      }

      if (character === ';' && call.length > 0) {
        tree.push(call);
        call = [];
      }
    }
  }

  return tree;
}

/**
 * Compiles a script against a given list of libraries into executable code
 * @param {string} script
 * @param {string[]} libraries
 * @return {string}
 */
function compileScript(script, libraries) {
  return `
    "use strict";
    {
      const libraries = ${JSON.stringify(libraries)};

      const script = ${JSON.stringify(parseScript(script))};

      let imports = Object.create(null);
      for (let library of libraries)
        new Function("exports", library)(imports);

      for (let [name, ...args] of script)
      {
        if (Object.prototype.hasOwnProperty.call(imports, name))
        {
          let value = imports[name];
          if (typeof value === "function")
            value(...args);
        }
      }
    }
  `;
}

function getExecutableCode(script) {
  let code = executableCode.get(script);
  if (code) {
    return code;
  }

  code = compileScript(script, [snippetsLibrarySource]);

  executableCode.set(script, code);
  return code;
}

// execute the specified snippet script on the tab / frame
function executeScript(script, tabId, frameId) {
  try {
    const details = {
      code: getExecutableCode(script),
      matchAboutBlank: true,
      runAt: 'document_start',
    };

    // Chrome <50 throws an exception if chrome.tabs.executeScript is called
    // with a frameId of 0.
    // Edge doesn't support frameId currently (as of 10/19/18)
    /*if (frameId !== 0) {
      details.frameId = frameId;
    }*/

    chrome.tabs.executeScript(tabId, details)
      .catch((error) => {
      // Sometimes a frame is added and removed very quickly, in such cases we
      // simply ignore the error.
        if (error.message === 'The frame was removed.') {
          return;
        }

        // Sometimes the frame in question is just not found. We don't know why
        // this is exactly, but we simply ignore the error.
        if (/^No frame with id \d+ in tab \d+\.$/.test(error.message)) {
          return;
        }

        throw error;
      });
  } catch (error) {
    // See the comment in the catch block associated with the call to
    // tabs.insertCSS for why we catch and ignore any errors here.
  }
}

fetch(chrome.extension.getURL('/snippets.js'), { cache: 'no-cache' })
  .then(response => (response.ok ? response.text() : ''))
  .then((text) => {
    snippetsLibrarySource = text;
  });
